/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/************************************************************
 * project=tweety                                           *
 * codename= kleiner EIBNET/IP server                       *
 * version=1.1                                              *
 ************************************************************/


/**
 * @file server.c
 *
 * This file contains tweety's main class method.
 *
 * @date created 16.05.05,
 *       last modified 02.09.05
 *
 * @author bernhard erb
 *         bernhard.erb@gmx.net
 *
 */ 

/*
 * ********************************************* includes ***
 */


#include "common.h"


/*
 * ******************************************* prototypes ***
 */


void *receiveUdp();		/* forward declaration: method in receive_udp.c */
void *receiveEib();		/* forward declaration: method in receive_eib.c */
void *sendUdp();		/* forward declaration: method in send_udp.c */
void *sendEib();		/* forward declaration: method in send_eib.c */
void *sendDisconnect_Request();	/* forward declaration: method in receive_udp.c*/
 
void server_SignalHandler(int sig); /* forward declaration: method in server.c*/


/*
 * ******************************************** functions ***
 */

/**
 * This method opens the KNX EIB device
 * and switches the BCU in Link Layer mode
 */

void openEIBDevice()
{

    if ((eib_fd = open(device, O_RDWR)) == -1) {	/* open file descriptor */
	char str[256];
	strncpy(str, "Error: can't open ", 256);
	strncat(str, device, 256);
	bailOut(str);
    }

    if (write(eib_fd, "\x46\x01\x00\x60\x12", 5) == -1) {	/* switch to link layer */
	bailOut("Error: can't reset bcu");
    }
}

/**
 * Closes the eib device descriptor.
 */
void closeEIBDevice() 
{

  if (eib_fd!=-1) {
    close(eib_fd);
    eib_fd=-1;
  }

}

/**
 * Sets the BCU table length to the
 * specified value.  
 */
void setTableLength(int tablelength)
{

  PC_GET_MESSAGE *reset = malloc(5);

  if (eib_fd == -1)
    openEIBDevice();
  reset->mc = 0x46;		/* pc_setval message */
  reset->length = 0x01;
  reset->address = 0x1601;	/* location of table size */
  *((byte *) reset + 4) = (byte) tablelength;
  
  write(eib_fd, (byte *) reset, 5);	/* reset table size value */
  write(eib_fd, "\x46\x01\x00\x60\xc0", 5);	/* reset bcu */
  
  
  free(reset);

  closeEIBDevice();
}

/**
 * This method returns the BCU
 * tablelength
 */

int readTableLength()
{
 
  int ret;
  int run = 1;
  PC_GET_MESSAGE *pc;
  byte *buffer = (byte *) malloc(32);
  
  signal(SIGALRM,server_SignalHandler);
  if (eib_fd == -1) {
    openEIBDevice();
  }
  
  if (write(eib_fd, "\x4c\x03\x01\x16", 4) == -1) {	/* request tablesize & address */
    bailOut("Error: can't write to bus");
  }
  
  while (run == 1) {		/* until we get the right answer */
      alarm(3);
      if ((read(eib_fd, buffer, 32)) == -1) {	/* read answer */
	bailOut("Error: can't receive packet from eib");
      }
      
      pc = (PC_GET_MESSAGE *) buffer;
      
      if (pc->mc == 0x4b && pc->address == 0x1601) {	/* the answer to our request */
	ret = *(buffer + 4);	/* tablesize is 1 bit after header */
	address = *((word *) (buffer + 5));	/* the next 2 bit are the address */
	run = 0;		/* stop while */
      }
  }
  alarm(0);

  closeEIBDevice();
  return (byte) ret;
}

/**
 * This method resets the BCU table size
 * closes the BCU driver device, 
 * and resets then the eib_fd
 * to -1 to mark it closed. Then it destroys 
 * the semaphor and unlinkes send and
 * receive pipe
 */
void freeResources()
{			
  
/* cleanup all needed ressources */
    if (eib_fd != -1) {
	setTableLength(tablesize);
	close(eib_fd);
	eib_fd = -1;           /* mark it closed*/
    }
    if (semid != (sem_t *) NULL) {	/* close semaphor */
	sem_destroy(semid);
	semid = (sem_t *) NULL;
    }

    unlink(SEND_PIPE);		/* remove pipes */
    unlink(RECEIVE_PIPE);
}

/**
 * This method acts as signal handler.
 * On SIGINT and SIGTERM the server is
 * halted. Before halting the server 
 * tries to reset the tablelength to
 * the original value.
 * SIGALRM is called by the alarm system call
 * which signals a that a BCU doesn't answer
 * to a read request.
 */
void server_SignalHandler(int sig)
{
    switch (sig) {
    case SIGINT:		/* shut down server */
	/*fall through */
    case SIGTERM:
	fprintf(stdout, "*** shutting down EIBNET/IP server ***\n");
	if (sock != -1)
	    sendDisconnect_Request();	/* disconnect client */

	pthread_kill(thread_receive_eib, SIGUSR1);	/* kill all threads */
	pthread_kill(thread_send_eib, SIGUSR1);

	kill(pid_SendUdp, SIGINT);	/* kill child processs */
	freeResources();
	fprintf(stdout, "server halted\n");
	exit(0);
	break;
    
    case SIGALRM:   
      fprintf(stdout,"BCU doesn't answer\n");
      close(eib_fd);
      exit(-1);
      break;
    }
}

/**
 * This method encapsulates all
 * general init procedures. 
 * - open eib device
 * - open socket
 * - reset address table length to 0
 * - init semaphor
 * - init shared memory
 * - start threads and fork process
 */
void initServer(void)
{				/* init all needed ressources */
    static struct ip_mreq command;
    int loop = 1;

    if (eib_fd == -1) {
	openEIBDevice();
    }

    /*
     * **** init socket ***
     */

    if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {	/* open socket */
	bailOut("Error: can't open socket");
    }


    serverAddr.sin_family = AF_INET;
    serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);	/* bind to any local address */
    serverAddr.sin_port = htons(serverport);	/* bind to SERVERPORT in header.h */


    if (bind(sock,		/* bind socket */
	     (struct sockaddr *) &serverAddr, sizeof(serverAddr)) < 0) {
	bailOut("Error: can't bind socket");
    }

    setsockopt(sock,	/*set multicast loop */
		   IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop));
    

    command.imr_multiaddr.s_addr = inet_addr(MULTICASTADDRESS);
    command.imr_interface.s_addr = htonl(INADDR_ANY);

    setsockopt(sock,	/* register multicast address */
		   IPPROTO_IP,
		   IP_ADD_MEMBERSHIP, &command, sizeof(command));



    /*
     * *** init eib device ***
     */

    tablesize = readTableLength();
    setTableLength(0);		/*set tablelength to 0 */

    /*
     * *** init pipes ***
     */


    if (mknod(SEND_PIPE, S_IFIFO | PERM, 0) < 0) {	/* make send pipe */
	bailOut("Error: can't construct pipe_send");
    }
    if (mknod(RECEIVE_PIPE, S_IFIFO | PERM, 0) < 0) {	/* make receive pipe */
	bailOut("Error: can't construct pipe_receive");
    }


    /*
     * *** init semaphor ***
     */


    if ((semid =
	 sem_open(KEY_SEMAPHOR, O_CREAT, PERM, 2)) == (sem_t *) NULL) {
	bailOut("Error: can't grab semaphor");
    }
    sem_post(semid);


    /*
     * *** init shared memory ***
     */


    if ((nSharedMem = shmget(KEY_SMEMORY,
			     sizeof(connection), 0666 | IPC_CREAT)) == -1)
    {
	bailOut("Error: can't obtain shared memory");
    }


    /*
     * *** fork and start threads ***
     */

    pid_ReceiveUdp = getpid();	/* receive_udp is the father process */

    switch (pid_SendUdp = fork()) {	/* send_udp is child process */
    case -1:
	bailOut("Error: can't fork process");
	break;
    default:
      openEIBDevice();
      /* start receive_eib as thread */
      iret_receive_eib =
	pthread_create(&thread_receive_eib, NULL, receiveEib,
		       (char *) NULL);
      
      
      /* start send_eib as thread */
      iret_send_eib = pthread_create(&thread_send_eib,
				     NULL, sendEib, (char *) NULL);
      
      signal(SIGINT, server_SignalHandler);	/* capture CTRL-C */
      signal(SIGTERM, server_SignalHandler);
      
      fprintf(stdout, "EIBNET/IP server started\n\n");
      
      receiveUdp();		/* start receive udp */
      break;
    case 0:
      sendUdp();		/* start send udp as child */
      break;
      
    }
    
}


/**
 * This method prints out the
 * usage and a short description
 * of available options
 */
void usage()
{
    fprintf(stdout, "*** usage ***\n");
    fprintf(stdout,
	    "Tweety is a small KNX EIBnet/IP server which allows you to connect to your EIB/KNX building automation system using EIBnet./IP\n If you are looking for an EIBnet/IP client please consider Calimero, which can be found at http://calimero.sourceforge.net .\n\n");
    fprintf(stdout, "Usage: tweety [-h,[-d DEVICE] [[-p Port][-n Name],[-r],[-w tablesize]]\n\n");  
    fprintf(stdout,"[-d device] The name of the tuwien auto PEI 16 driver device. By default /dev/eib is used.\n"); 
    fprintf(stdout,"[-p port] Causes tweety to use the specified port instead of default port 3671. Note that clients device discovery will only work with the default port.\n");
    fprintf(stdout,"[-n name] Sets the device name, that is processed on each description request. If the option is not set the projects codename \"Tweety 1.1\" is used by default.\n");
    fprintf(stdout,"[-r] This option is used to retrieve the BCU address table size.\n");
    fprintf(stdout,"[-w tablesize] Sets the BCU address table size to the given value.\n\n"); 
    fprintf(stdout,"[-h] Prints this usage\n\n");

    exit(-1);			/* exit with error code */
}


/**
 * Main method. Inits needed
 * variables, examines options
 * and  starts relevant
 * functions
 */
int main(int argc, char **argv)
{
    int service = 0;
    int c = 0;

    szCommand = argv[0];
    serverport = SERVERPORT;
    tablesize = 0;
    friendlyName = "Tweety 1.1";
    device = STANDARDDEVICE;
    eib_fd = -1;      /* mark eib_fd as closed */

    /*
     * *** print init *** 
     */

    fprintf(stdout, "************************************\n");
    fprintf(stdout, "Tweety Version 1.1\n");
    fprintf(stdout, "auto.tuwien.ac.at/KNX\n");
    fprintf(stdout, "tweety -h for usage\n\n");
    fprintf(stdout,
	    "Copyright (C) 2005 Bernhard Erb\n*This program is free software; you can redistribute it and/or\n*modify it under the terms of the GNU General Public License\n*as published by the Free Software Foundation; either version 2\n*of the License, or (at your option) any later version.\n*");
    fprintf(stdout,
	    "This program is distributed in the hope that it will be useful,\n*but WITHOUT ANY WARRANTY; without even the implied warranty of\n*MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n*GNU General Public License for more details.\n\n");


    /*
    * *** reading options ***
    */ 
    while ((c = getopt(argc, argv, "d:p:r::w:n:h::")) != EOF)
	switch (c) {
	case 'r':
	    if (service != 0)	/* if just another service was requested */
		usage();
	    service = 1;
	    break;
	case 'n':
	    friendlyName = optarg;
	    break;
	case 'w':
	    if (service != 0)
		usage();
	    service = 2;
	    tablesize = (byte) strtol(optarg, NULL, 10);
	    break;
	case 'd':
	    device = optarg;	/*device is optarg */
	    break;
	case 'p':
	    serverport = strtol(optarg, NULL, 10);
	    break;
	case 'h':
	default:		/*unknown option */ 
	  usage();
	    break;
	}


    if (service == 0) {		/*start the server */
	fprintf(stdout, "*** starting EIBNET/IP server ***\n");
	fprintf(stdout, "Connecting with device %s \n",device);
	fprintf(stdout, "Friendly name is %s \n",friendlyName);
	fprintf(stdout, "The %s address table length is %d\n", device,
		readTableLength());
	initServer();
    } else if (service == 1) {	/* read the bcu table length */
	fprintf(stdout, "*** reading address table length ***\n");
	fprintf(stdout, "The %s address table length is %d\n", device,
		readTableLength());
	close(eib_fd);
    } else if (service == 2) {	/*set the bcu table length */
	fprintf(stdout, "*** writing address table length ***\n");
	fprintf(stdout, "Current %s address table length is %d\n", device,
		readTableLength());
	fprintf(stdout, "Setting the address table length\n");
	setTableLength(tablesize);
	fprintf(stdout, "The %s address table length is now %d\n", device,
		readTableLength());
	close(eib_fd);
    }

    return 0;
}

/*
 * ********************************************* EOF ***
 */
